using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// BeatIt v1.1
// Modified by HAL9001,
// based on hamzhox VAMAudioAnalyzer
//
// Allows precise control of triggers based on audio frequencies bands.
// Does not care bout audio source overall volume as it auto-adjusts each band.
//
// v1.0 - 2023/10/16 - Initial Version
// v1.1 - 2024/04/28 - Modified AudioSource acquisition - properly finds Person HeadAudio now as well
//

namespace HAL9001
{
    public class BeatIt : MVRScript
    {
		static readonly int     NUM_BANDS           = 8;
		static readonly float   AUDIO_SAMPLE_RATE   = 48000f;
		static readonly int     FFT_BUFFER_SIZE     = 1024;  // should capture about 1/40h of a second @ 48000 Hz samplerate
		static readonly float[] DEFAULT_FREQUENCIES = new[] {50f, 100f, 250f, 500f, 1000f, 2500f, 5000f, 10000f};
		

		// JSONStorables
		public  JSONStorableFloat m_RiseSpeed;
		public  JSONStorableFloat m_FallSpeed;
		public  JSONStorableFloat m_FallAcceleration;
		public  JSONStorableFloat m_VolumeCompensation;
		public  JSONStorableStringChooser m_bandToEdit;
		
		// Game Objects and components
		protected AudioSource m_audioSource;
		
		// Variables
		protected bool    m_hasSignal;
		protected float[] m_spectrum;
		protected List<FrequencyBand>  m_bands;
		


		protected class FrequencyBand
		{
			protected BeatIt             outer;
			protected int                bandIdx;
			protected string             baseName;
			protected JSONStorableString bandName;
			protected JSONStorableFloat  centerFrequency;
			protected JSONStorableFloat  bandwidth;
			protected JSONStorableFloat  currentLevel;
			public    EventTrigger       trigger;
			protected UIDynamicButton    btnTrigger;
			protected UIDynamic          spacer1, spacer2;

			protected float m_level, m_maxLevel, m_animatedLevel, m_levelFall;
			

			public FrequencyBand(BeatIt _outer, int _bandIdx, float _defaultFrequency)
			{
				outer    = _outer;
				bandIdx  = _bandIdx;
				baseName = "Band " + bandIdx;

				bandName        = new JSONStorableString(baseName + " Name",        baseName); 
				centerFrequency = new JSONStorableFloat( baseName + " Frequency",   _defaultFrequency,   (float val) => {}, 20f, 20000f, true, true);
				bandwidth       = new JSONStorableFloat( baseName + " Bandwidth",   _defaultFrequency/2, (float val) => {}, 10f, 10000f, true, true);

				currentLevel = new JSONStorableFloat( baseName + " Level", 0f, (float val) => {},  0f, 1f);
				currentLevel.isStorable   = false;
				currentLevel.isRestorable = false;
				
				trigger = new EventTrigger(outer, baseName + " Actions");
				trigger.active = true;

				m_level         = 0;
				m_maxLevel      = 0;
				m_animatedLevel = 0;
				m_levelFall     = 0;
			}


			public string GetName()
			{
				return bandName.val;
			}


			public void CreateUI()
			{
				UIDynamicSlider slider;
				if (GetName() == outer.m_bandToEdit.val)
				{
					if (bandIdx > 1) { spacer1 = outer.CreateSpacer(true); spacer1.height = 20; }

					slider = outer.CreateSlider(currentLevel, true);
					// outer.CreateTextField(bandName, true);
					btnTrigger = outer.CreateButton(baseName + " Triggers", true);
					btnTrigger.button.onClick.AddListener( () => {
						trigger.OpenPanelTransition();
					});

					var sldF = outer.CreateSlider(centerFrequency, true);
					sldF.ConfigureQuickButtons(-1, -10, -100, -1000, 1, 10, 100, 1000);
					var sldBW = outer.CreateSlider(bandwidth, true);
					sldBW.ConfigureQuickButtons(-1, -10, -100, -1000, 1, 10, 100, 1000);


					if (bandIdx < NUM_BANDS) { spacer2 = outer.CreateSpacer(true); spacer2.height = 20; }
				}
				else
				{
					slider = outer.CreateSlider(currentLevel, true);
					slider.labelText.text = String.Format("{0} ({1} ± {2})",
						bandName.val,
						FrequencyToString(centerFrequency.val),
						FrequencyToString(bandwidth.val)
					);
				}

				slider.quickButtonsEnabled  = false;
				slider.defaultButtonEnabled = false;
				slider.rangeAdjustEnabled   = false;
				slider.sliderControl.slider.interactable = false;
			}


			public void DestroyUI()
			{
				outer.RemoveSpacer(spacer1);
				outer.RemoveSlider(currentLevel);
				outer.RemoveButton(btnTrigger);
				outer.RemoveSlider(centerFrequency);
				outer.RemoveSlider(bandwidth);
				outer.RemoveSpacer(spacer2);
			}


			public bool ProcessSpectrum(float[] _spectrum)
			{
				int   startIdx = GetIndexFromFrequency(centerFrequency.val - bandwidth.val / 2f);
				int   endIdx   = GetIndexFromFrequency(centerFrequency.val + bandwidth.val / 2f);

				m_level = 0;
				for (int i = startIdx; i <= endIdx; i++) {
					m_level += _spectrum[i];
				}
				m_level /= (endIdx - startIdx + 1);
				
				if (outer.m_audioSource != null)
				{
					// compensate by source volume
					m_level /= Mathf.Max(0.01f, outer.m_audioSource.volume); // 0.01: avoid divide by zero
				}

				return m_level > 0.0001f;
			}

			
			public void Animate() 
			{
				m_maxLevel = Mathf.Max(m_level, m_maxLevel);
				
				float level = outer.m_hasSignal ? (m_level / m_maxLevel) : 0;
				if (outer.m_audioSource != null) 
				{
					level *= 1f - ((1f - outer.m_VolumeCompensation.val) * (1f - outer.m_audioSource.volume));
				}

				if ( level > m_animatedLevel ) {
					m_animatedLevel  += outer.m_RiseSpeed.val * (level - m_animatedLevel);
					m_levelFall       = outer.m_FallSpeed.val;
				}
				else
				{
					m_animatedLevel = Mathf.Max(0f, m_animatedLevel - m_levelFall * Time.deltaTime);
					m_levelFall    *= outer.m_FallAcceleration.val;
				}

				currentLevel.val = m_animatedLevel;
				trigger.TriggerTransition(m_animatedLevel);
				trigger.Update();
			}


			public void ResetCalibration()
			{
				m_maxLevel = 0.001f;
			}


			protected int GetIndexFromFrequency(float _frequency)
			{
  				int bin = (int) (_frequency / ((AUDIO_SAMPLE_RATE / 2f) / FFT_BUFFER_SIZE));
				if (bin <  0)               { bin = 0; }
				if (bin >= FFT_BUFFER_SIZE) { bin = FFT_BUFFER_SIZE-1; }
				return bin;
			}

			protected string FrequencyToString(float _frequency)
			{
  				if (_frequency < 1000.0f) { return String.Format("{0:F0}Hz",  _frequency); }
				else                      { return String.Format("{0:F0}kHz", _frequency/1000.0f); }
			}
		}


		public override void Init()
        {			 
            try
            {
				// Grabbing audio source
				m_audioSource = containingAtom.GetComponentInChildren(typeof(AudioSource)) as AudioSource;
				if (m_audioSource == null)
				{
					m_audioSource = containingAtom.reParentObject.GetComponentInChildren(typeof(AudioSource)) as AudioSource;
				}				
				if (m_audioSource == null)
				{
					logInfo("Could not find direct AudioSource - Using scene AudioListener");
				}				
				
				m_hasSignal = false;
				m_spectrum  = new float[FFT_BUFFER_SIZE];

				// Loading triggers assets
				SimpleTriggerHandler.LoadAssets();

				m_bands = new List<FrequencyBand>();
				for (int i=0; i<NUM_BANDS; i++)
				{
					FrequencyBand b = new FrequencyBand(this, i+1, DEFAULT_FREQUENCIES[i]);
					m_bands.Add(b);
				}

				CreateUI();
			}
            catch (Exception e)
            {
                logError("Exception caught: " + e);
            }
		}
		

		protected void CreateUI()
		{
			/////////////////////////////
			// Left Side
			
			createStaticDescriptionText("Parameters","<color=#000><size=32><b>Parameters</b></size></color>", false, 40);
			
			m_RiseSpeed = new JSONStorableFloat("Rise Ratio", 1f, 0f, 1f);
			RegisterFloat(m_RiseSpeed);
			var sliderRS = CreateSlider(m_RiseSpeed);
			sliderRS.valueFormat = "F4";

			m_FallSpeed = new JSONStorableFloat("Fall Speed", 0.1f, 0.0001f, 1f);
			RegisterFloat(m_FallSpeed);
			var sliderFS = CreateSlider(m_FallSpeed);
			sliderFS.valueFormat = "F4";

			m_FallAcceleration = new JSONStorableFloat("Fall Acceleration", 1.2f, 1.001f, 2f);
			RegisterFloat(m_FallAcceleration);
			var sliderFA = CreateSlider(m_FallAcceleration);
			sliderFA.valueFormat = "F3";
			
			if (m_audioSource != null)
			{
				m_VolumeCompensation = new JSONStorableFloat("Volume Compensation", 1f, 0f, 1f);
				RegisterFloat(m_VolumeCompensation);
				var slider = CreateSlider(m_VolumeCompensation);
			}	
			
			CreateSpacer();

			List<string> bandChoices = new List<string>();
			bandChoices.Add("None");
			foreach (FrequencyBand b in m_bands) { bandChoices.Add(b.GetName()); }
			m_bandToEdit = new JSONStorableStringChooser("Band to Edit", bandChoices, bandChoices[0], "Band to Edit");
			m_bandToEdit.setCallbackFunction += BandToEditChanged;
			var chooserBtE = CreateScrollablePopup(m_bandToEdit, false);

			/////////////////////////////
			// Right Side
			
			createStaticDescriptionText("Frequency Triggers","<color=#000><size=32><b>Frequency Triggers</b></size></color>", true, 40);

			foreach (FrequencyBand b in m_bands)
			{
				b.CreateUI();
			}
		}


		protected void BandToEditChanged(string bandName)
		{
			foreach (FrequencyBand b in m_bands)
			{
				b.DestroyUI();
			}
			foreach (FrequencyBand b in m_bands)
			{
				b.CreateUI();
			}
		}


		public void Update() 
		{
			if (SuperController.singleton.freezeAnimation) return;
			
			// Updating the spectrum itself
			GetAudioSpectrum();

			m_hasSignal = false;
			foreach (FrequencyBand b in m_bands)
			{
				m_hasSignal |= b.ProcessSpectrum(m_spectrum);
			}
			if ((m_audioSource != null) && (!m_audioSource.isPlaying)) 
			{
				 m_hasSignal = false; 
			}
			if (!m_hasSignal)
			{
				foreach (FrequencyBand b in m_bands)
				{
					b.ResetCalibration();
				}
			}

			foreach (FrequencyBand b in m_bands)
			{
				b.Animate();
			}
		}
		

		protected void GetAudioSpectrum() 
		{
			if (m_audioSource != null) {
				m_audioSource.GetSpectrumData(m_spectrum, 0, FFTWindow.Blackman);
			} else {
				AudioListener.GetSpectrumData(m_spectrum, 0, FFTWindow.Blackman);
			}
		}		
		

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				foreach (FrequencyBand b in m_bands) {
					jc[b.trigger.Name] = b.trigger.GetJSON(base.subScenePrefix);
				}				
			}
			return jc;
		}


		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				foreach (FrequencyBand b in m_bands) {
					b.trigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				}
			}
		}
			

		protected static void logInfo(string text) 
		{
			SuperController.LogMessage($"BeatIt: {text}");
		}
		

		protected static void logError(string text) 
		{
			SuperController.LogError($"BeatIt: {text}");
		}
		

		protected void disableScrollOnText(UIDynamicTextField target)
		{
			ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
			if( targetSR != null ) {
				targetSR.horizontal = false;
				targetSR.vertical = false;
			}
		}
		

		protected void createStaticDescriptionText(string DescTitle, string DescText, bool rightSide, int fieldHeight) 
		{
			JSONStorableString staticDescString = new JSONStorableString(DescTitle,DescText);
			UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
			staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
			LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
			sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
			staticDescStringField.height = fieldHeight;
			disableScrollOnText(staticDescStringField);
		}
				
		public void OnDestroy() 
		{
			foreach (FrequencyBand b in m_bands) 
			{
				b.trigger.Remove();
			}
		}


		// ***********************************************************
		// EXTERNAL TOOLS - Thank you great coders for your content!
		// ***********************************************************
		
		// *********** MacGruber_Utils.cs START *********************
		
		// ===========================================================================================
		// TriggerHandler implementation for easier handling of custom triggers.
		// Essentially call this in your plugin init code:
		//     StartCoroutine(SimpleTriggerHandler.LoadAssets());
		//
		// Credit to AcidBubbles for figuring out how to do custom triggers.
		//
		// hazmhox note : Edited a bit to allow the use of transitions ( original code coming from LogicBricks 4 )
		public class SimpleTriggerHandler : TriggerHandler
		{
			public static bool Loaded { get; private set; }

			private static SimpleTriggerHandler myInstance;

			private RectTransform myTriggerActionsPrefab;
			private RectTransform myTriggerActionMiniPrefab;
			private RectTransform myTriggerActionDiscretePrefab;
			private RectTransform myTriggerActionTransitionPrefab;

			public static SimpleTriggerHandler Instance {
				get {
					if (myInstance == null)
						myInstance = new SimpleTriggerHandler();
					return myInstance;
				}
			}

			public static void LoadAssets()
			{
				SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
			}

			private IEnumerator LoadAssetsInternal()
			{
				foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
					yield return x;

				Loaded = true;
			}

			private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
			{
				AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
				if (request == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
				yield return request;
				GameObject go = request.GetAsset<GameObject>();
				if (go == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
				RectTransform prefab = go.GetComponent<RectTransform>();
				if (prefab == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
				assign(prefab);
			}


			void TriggerHandler.RemoveTrigger(Trigger t)
			{
				// nothing to do
			}

			void TriggerHandler.DuplicateTrigger(Trigger t)
			{
				throw new NotImplementedException();
			}

			RectTransform TriggerHandler.CreateTriggerActionsUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionMiniUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
			}

			void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
			{
				UnityEngine.Object.Destroy(rt?.gameObject);
			}
		}

		// Wrapper for easier handling of custom triggers.
		public class EventTrigger : Trigger
		{
			public string Name {
				get { return name; }
				set { name = value; myNeedInit = true; }
			}

			public MVRScript Owner {
				get; private set;
			}

			private string name;
			private bool myNeedInit = true;

			public EventTrigger(MVRScript owner, string name)
			{
				Name = name;
				Owner = owner;
				handler = SimpleTriggerHandler.Instance;
			}

			public void OpenPanelActionStart()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					logError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
			
			public void OpenPanelTransition()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					logError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				
				//getRoot(triggerActionsParent,0);
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab2/Label").GetComponent<Text>().text = "Frequencies Actions";
					triggerActionsPanel.Find("Content/Tab1").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(true);
					triggerActionsPanel.Find("Content/Tab2").GetComponent<Toggle>().isOn = true;					
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
	
			public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
			{
				if (jc.HasKey(Name))
				{
					JSONClass tc = jc[Name].AsObject;
					if (tc != null)
						base.RestoreFromJSON(tc, subScenePrefix, isMerge);
				}
				else if (setMissingToDefault)
				{
					base.RestoreFromJSON(new JSONClass());
				}
			}

			public void Trigger()
			{
				active = true;
				active = false;
			}
			
			public void TriggerTransition(float val)
			{
				this.transitionInterpValue = val;
			}
		}
		// *********** MacGruber_Utils.cs END *********************
	
	}
}